#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.json_format import MessageToJson
from prometheus_client import Counter, Histogram
from th2_grpc_common.common_pb2 import EventBatch

import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_subscriber import AbstractRabbitSubscriber
from th2_common.schema.util.util import get_debug_string_event


class EventBatchSubscriber(AbstractRabbitSubscriber):
    INCOMING_EVENT_BATCH_QUANTITY = Counter('th2_mq_incoming_event_batch_quantity',
                                            'Quantity of incoming event batches')
    INCOMING_EVENT_QUANTITY = Counter('th2_mq_incoming_event_quantity',
                                      'Quantity of incoming events')
    EVENT_PROCESSING_TIME = Histogram('th2_mq_event_processing_time',
                                      'Time of processing events',
                                      buckets=common_metrics.DEFAULT_BUCKETS)

    def get_delivery_counter(self) -> Counter:
        return self.INCOMING_EVENT_BATCH_QUANTITY

    def get_content_counter(self) -> Counter:
        return self.INCOMING_EVENT_QUANTITY

    def get_processing_timer(self) -> Histogram:
        return self.EVENT_PROCESSING_TIME

    def extract_count_from(self, batch: EventBatch):
        return len(self.get_events(batch))

    def get_events(self, batch: EventBatch) -> list:
        return batch.events

    @staticmethod
    def value_from_bytes(body):
        event_batch = EventBatch()
        event_batch.ParseFromString(body)
        return [event_batch]

    def filter(self, value) -> bool:
        return True

    def extract_labels(self, batch):
        return []

    def to_trace_string(self, value):
        return MessageToJson(value)

    def to_debug_string(self, value):
        return get_debug_string_event(value)



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.json_format import MessageToJson
from prometheus_client import Counter, Histogram
from th2_grpc_common.common_pb2 import MessageGroupBatch

import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_batch_subscriber import AbstractRabbitBatchSubscriber, \
    Metadata
from th2_common.schema.util.util import get_debug_string_group, get_session_alias_and_direction_group


class RabbitMessageGroupBatchSubscriber(AbstractRabbitBatchSubscriber):
    INCOMING_MSG_GROUP_BATCH_QUANTITY = Counter('th2_mq_incoming_msg_group_batch_quantity',
                                                'Quantity of incoming message group batches',
                                                common_metrics.DEFAULT_LABELS)
    INCOMING_MSG_GROUP_QUANTITY = Counter('th2_mq_incoming_msg_group_quantity',
                                          'Quantity of incoming message groups',
                                          common_metrics.DEFAULT_LABELS)
    MSG_GROUP_PROCESSING_TIME = Histogram('th2_mq_msg_group_processing_time',
                                          'Time of processing message groups',
                                          buckets=common_metrics.DEFAULT_BUCKETS)

    def get_delivery_counter(self) -> Counter:
        return self.INCOMING_MSG_GROUP_BATCH_QUANTITY

    def get_content_counter(self) -> Counter:
        return self.INCOMING_MSG_GROUP_QUANTITY

    def get_processing_timer(self) -> Histogram:
        return self.MSG_GROUP_PROCESSING_TIME

    def extract_count_from(self, batch: MessageGroupBatch):
        return len(batch.groups)

    def get_messages(self, batch) -> list:
        return batch.groups

    def extract_metadata(self, message) -> Metadata:
        raise ValueError

    @staticmethod
    def value_from_bytes(body):
        message_group_batch = MessageGroupBatch()
        message_group_batch.ParseFromString(body)
        return [message_group_batch]

    def extract_labels(self, batch):
        return get_session_alias_and_direction_group(self.get_messages(batch)[0].messages[0])

    def to_trace_string(self, value):
        return MessageToJson(value)

    def to_debug_string(self, value):
        return get_debug_string_group(value)



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from abc import ABC, abstractmethod

from prometheus_client import Counter

from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.impl.rabbitmq.connection.reconnecting_publisher import ReconnectingPublisher
from th2_common.schema.message.message_sender import MessageSender

logger = logging.getLogger(__name__)


class AbstractRabbitSender(MessageSender, ABC):

    def __init__(self, connection_manager: ConnectionManager, exchange_name: str, send_queue: str) -> None:
        self.__publisher: ReconnectingPublisher = connection_manager.publisher
        self.__exchange_name: str = exchange_name
        self.__send_queue: str = send_queue
        self.__closed = True

    def start(self):
        if self.__send_queue is None or self.__exchange_name is None:
            raise Exception('Sender can not start. Sender did not init')
        self.__closed = False

    def is_close(self) -> bool:
        return self.__closed

    def close(self):
        self.__closed = True

    def send(self, message):
        if message is None:
            raise ValueError('Value for send can not be null')
        try:
            self.__publisher.publish_message(exchange_name=self.__exchange_name,
                                             routing_key=self.__send_queue,
                                             message=self.value_to_bytes(message))

            if logger.isEnabledFor(logging.TRACE):
                logger.trace(f'Sending to exchange_name = "{self.__exchange_name}", '
                             f'routing_key = "{self.__send_queue}", '
                             f'message = {self.to_trace_string(message)}')
            elif logger.isEnabledFor(logging.DEBUG):
                logger.debug(f'Sending to exchange_name = "{self.__exchange_name}", '
                             f'routing_key = "{self.__send_queue}", '
                             f'message = {self.to_debug_string(message)}')

            counter = self.get_delivery_counter()
            counter.inc()
            content_counter = self.get_content_counter()
            content_counter.inc(self.extract_count_from(message))

        except Exception:
            logger.exception('Can not send')

    @staticmethod
    @abstractmethod
    def value_to_bytes(value):
        pass

    @abstractmethod
    def get_delivery_counter(self) -> Counter:
        pass

    @abstractmethod
    def get_content_counter(self) -> Counter:
        pass

    @abstractmethod
    def extract_count_from(self, batch):
        pass

    @abstractmethod
    def to_trace_string(self, value):
        pass

    @abstractmethod
    def to_debug_string(self, value):
        pass



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from th2_common.schema.filter.strategy.impl.default_filter_strategy import DefaultFilterStrategy
from th2_common.schema.message.configuration.message_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_queue import AbstractRabbitQueue
from th2_common.schema.message.impl.rabbitmq.configuration.subscribe_target import SubscribeTarget
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.impl.rabbitmq.group.rabbit_message_group_batch_sender import \
    RabbitMessageGroupBatchSender
from th2_common.schema.message.impl.rabbitmq.group.rabbit_message_group_batch_subscriber import \
    RabbitMessageGroupBatchSubscriber
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber


class RabbitMessageGroupBatchQueue(AbstractRabbitQueue):

    def create_sender(self, connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration) -> MessageSender:
        return RabbitMessageGroupBatchSender(connection_manager, queue_configuration.exchange,
                                             queue_configuration.routing_key)

    def create_subscriber(self, connection_manager: ConnectionManager,
                          queue_configuration: QueueConfiguration) -> MessageSubscriber:
        subscribe_target = SubscribeTarget(queue_configuration.queue, queue_configuration.routing_key)
        return RabbitMessageGroupBatchSubscriber(connection_manager,
                                                 queue_configuration,
                                                 DefaultFilterStrategy(),
                                                 subscribe_target)



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.json_format import MessageToJson
from prometheus_client import Counter, Histogram
from th2_grpc_common.common_pb2 import MessageBatch, MessageGroupBatch

import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_batch_subscriber import AbstractRabbitBatchSubscriber, \
    Metadata
from th2_common.schema.util.util import get_debug_string, get_session_alias_and_direction


class RabbitParsedBatchSubscriber(AbstractRabbitBatchSubscriber):
    INCOMING_PARSED_MSG_BATCH_QUANTITY = Counter('th2_mq_incoming_parsed_msg_batch_quantity',
                                                 'Quantity of incoming parsed message batches',
                                                 common_metrics.DEFAULT_LABELS)
    INCOMING_PARSED_MSG_QUANTITY = Counter('th2_mq_incoming_parsed_msg_quantity',
                                           'Quantity of incoming parsed messages',
                                           common_metrics.DEFAULT_LABELS)
    PARSED_MSG_PROCESSING_TIME = Histogram('th2_mq_parsed_msg_processing_time',
                                           'Time of processing parsed messages',
                                           buckets=common_metrics.DEFAULT_BUCKETS)

    def get_delivery_counter(self) -> Counter:
        return self.INCOMING_PARSED_MSG_BATCH_QUANTITY

    def get_content_counter(self) -> Counter:
        return self.INCOMING_PARSED_MSG_QUANTITY

    def get_processing_timer(self) -> Histogram:
        return self.PARSED_MSG_PROCESSING_TIME

    def extract_count_from(self, batch: MessageBatch):
        return len(self.get_messages(batch))

    def get_messages(self, batch: MessageBatch) -> list:
        return batch.messages

    def extract_metadata(self, message) -> Metadata:
        metadata = message.metadata
        return Metadata(message_type=metadata.message_type,
                        direction=metadata.id.direction,
                        sequence=metadata.id.sequence,
                        session_alias=metadata.id.connection_id.session_alias)

    @staticmethod
    def value_from_bytes(body):
        message_group_batch = MessageGroupBatch()
        message_group_batch.ParseFromString(body)

        message_batches = []
        for message_group in message_group_batch.groups:
            messages = []
            for any_message in message_group.messages:
                any_message.HasField('message')
                messages.append(any_message.message)
            message_batches.append(MessageBatch(messages=messages))

        return message_batches

    def extract_labels(self, batch):
        return get_session_alias_and_direction(self.get_messages(batch)[0].metadata.id)

    def to_trace_string(self, value):
        return MessageToJson(value)

    def to_debug_string(self, value):
        return get_debug_string(self.__class__.__name__, [message.metadata.id for message in self.get_messages(value)])



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.json_format import MessageToJson
from prometheus_client import Counter
from th2_grpc_common.common_pb2 import MessageGroupBatch

from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_sender import AbstractRabbitSender
from th2_common.schema.util.util import get_debug_string_group


class RabbitMessageGroupBatchSender(AbstractRabbitSender):
    OUTGOING_MSG_GROUP_BATCH_QUANTITY = Counter('th2_mq_outgoing_msg_group_batch_quantity',
                                                'Quantity of outgoing message group batches')
    OUTGOING_MSG_GROUP_QUANTITY = Counter('th2_mq_outgoing_msg_group_quantity',
                                          'Quantity of outgoing message groups')

    def get_delivery_counter(self) -> Counter:
        return self.OUTGOING_MSG_GROUP_BATCH_QUANTITY

    def get_content_counter(self) -> Counter:
        return self.OUTGOING_MSG_GROUP_QUANTITY

    def extract_count_from(self, batch: MessageGroupBatch):
        return len(batch.groups)

    @staticmethod
    def value_to_bytes(value: MessageGroupBatch):
        return value.SerializeToString()

    def to_trace_string(self, value):
        return MessageToJson(value)

    def to_debug_string(self, value):
        return get_debug_string_group(value)
